\ FORGET.F

\ This is a modified version of FORGET.F for dual dictionaries,
\ by Stephen M. Brault.

\ Received: 05 Feb 96 05:26:38 EST
\     From: "Stephen M. Brault" <72163.3434@compuserve.com>

\ February 5th, 1996 - 22:43 tjz
\ Renamed _TRIM? to SYS-TRIM? which describes its function.
\ Renamed TRIM to FULL-TRIM, to avoid a name conflict with the original
\ simple version of TRIM which is used in other areas of the system, and
\ takes slightly different parameters.

\ Notice that TRIM? is defered, so its function can be changed to make
\ FULL-TRIM trim things other than the dictionary if desired.

.( Loading Forget Wordset...)

\ FORGET limitations:
\ 1) You cannot directly forget an IN-SYSTEM word since
\ it would require extra work to determine the new DP.
\ You can indirectly forget an IN-SYSTEM word by forgetting
\ an earlier IN-APPLICATION word.

variable fence                          \ cannot forget below this address

: %trim?     ( nfa addr -- f )       \ TRUE if forgetting nfa removes addr from dict
    swap here between ;

defer trim? ' %trim? is trim?

: full-trim      ( nfa link -- )         \ Order-independent trim
        begin   ( nfa link^ )
                dup @
        while   ( nfa link^ )
                2dup @ trim?
                if      dup @ @ over !  \ Delete chain member
                else    @               \ Retain chain member
                then
        repeat  2drop ;

: trim-chains   ( nfa -- nfa )          \ trim down the chain linked list
        chain-link
        begin   @ ?dup
        while   2dup -2 cells+ full-trim
        repeat  dup chain-link full-trim ;

: trim-loadfile ( nfa -- nfa )
        dup loadfile full-trim ;

forget-chain chain-add trim-loadfile

comment:
: trim-defer    ( nfa -- nfa )          \ trim deferred word list to nfa
        dup defer-list full-trim
        defer-list
        begin   @ ?dup
        while   2dup cell- @
                trim?                   \ if forgetting IS word
                if      dup cell+ @     \ revert to default
                        over cell- !
                then
        repeat ;

forget-chain chain-add trim-defer
comment;

: vtrim ( nfa voc-thread -- ) \ trim VOC-THREAD back to nfa
        dup voc#threads 0
        do      2dup i cells+ full-trim
        loop    2drop ;

: (forget) ( cfa -- )   \ assumes count follows name
        >name
        dup fence @ 1- trim? abort" in protected dictionary"
        voc-link 2dup full-trim
        begin   @ ?dup
        while   2dup vlink>voc vtrim
        repeat
        forget-chain do-chain   \ do forget-chain before trimming it
        trim-chains
        context #vocs cells+ context
        do      dup i @ trim?
                if      [ ' forth vcfa>voc ] literal i !
                then
  cell +loop
        nfa-count drop dp ! \ update SYS-HERE
;


: forget ( -<name>- )
        bl word count
        current @ search-wordlist 0= ?missing ( cfa ) (forget) ;

: do-mark ( -- )        \ mark must redefine the name that was forgotten
        does>   { does-adr \ mark-name$ -- }
                MAXSTRING LocalAlloc: mark-name$
                s" mark " mark-name$ place
                does-adr body> dup >name nfa-count mark-name$ +place
                (forget) forth definitions
                mark-name$ count evaluate ;

: mark  ( -<name>- )
        create do-mark ;

: do-marker ( -- )
        does> @ (forget) forth definitions ;

: marker ( -<name>- ) ( ANS)
        create here body> , do-marker ;

here fence !

.( OK) cr