\ OOP package -- Martin Läuter

.( Loading OOP package... )

comment:
  This file implements my classes package. Some ideas I was missing from
other implementations were put in, others ignored. We stick with single
inheritance.
  First about syntax: Friends of Win32Forth will miss the "message: object"
syntax. But I missed the possibility passing methods on the stack, so
I switched to "object message". I liked the possibility of 'diving' into
objects, i.e. having access to nested instance variable the way SwiftForth
does it. But you cannot tick such an instance variable. To repair this
I had to extend the compiler which renders the approach not portable.
But now we can say:
   object.method             \ early binding
   object method             \ late binding
   object class.method       \ early binding with type cast
   object.ivar.method        \ early binding
   object.ivar method        \ late binding
   object.ivar class.method  \ early binding with type cast
And you can nest as many ivars as needed. If you don't know the class
of a nested ivar, use class.ivar.method for that.
That's the way it is thought.
  Now for declaration: Here is an example:
CLASS list SUPER> object
VAR first
VAR last
object wert
:M construct ." In List constructor" cr ;
Ns:M print ." List@" . ;
END-CLASS

You see the defining word CLASS which is paired by END-CLASS
The parent class is declared by SUPER>. Methods are defined using
:M which is paired by ;. I didn't use : for defining methods as in
other implementations because I lose the possibility of defining
private helper words, not supposed to be methods.
:M saves the object pointer which is reset on exit. You may write
NO-SAVE :M (shorter NS:M) when you don't want to save the object pointer
(which is tos then and not accessible by OP) for some speed improvement.
You may use PUBLIC:, PRIVATE: and PROTECTED: to hide methods and
instance variables from views from outside. I would like to have
static variables and methods, but this is not yet implemented.
  Every class should have a CONSTRUCT and a DESTRUCT method.
They are called automatically every time an object is constructed
resp. destructed (by NEW> or as an instance variable).

Now for some implementation details: we have an object pointer OP
which is used for accessing data. The cell before the object data
is the class pointer, so that late binding is possible. There is
a wordlist method-wl where all declared methods are stored (exactly
once). So they are hidden from outer inspection, but the uniqueness
makes it possible to send arbitrary messages to arbitrary objects
and still get the right result.

To Do List: implement missing features, think about how Be classes
may be imported and used like the Forth ones.
comment;

\ : cells+ cells + ;
\ : h. base @ hex swap u. base ! ;
\ : under+ rot + swap ;

\ variable op

vocabulary oop
only forth also oop definitions

0 value ^class

wordlist value method-wl

\ : link,
\     here  over @ ,  swap ! ;

my macro >sup " cell+ "
my macro >mfa " 2 cells+ "
my macro >ifa " 3 cells+ "
my macro >size " 4 cells+ "
my macro >wl " 5 cells+ "

variable class-link  0 class-link !

: method ( "name" -- meth-id )
    >in @  bl word count  method-wl search-wordlist
    if nip  >body
    else >in !  get-current >r
      method-wl set-current
      create
      r> set-current
      here
    then ;

variable find-opaque
variable save?
variable mth-xt
variable opaque
variable oldcur
variable inrec?
0 value obj-xt
0 value class-xt

defer ivinit    ' 2drop is ivinit

my macro mth>id " cell+ "
my macro mth>opq " 2 cells+ "
my macro mth>xt " 3 cells+ "

: find-mth ( meth-id class -- xt t | f )
    my >mfa
    begin @ dup while
      2dup my mth>id @ =
    until
      nip  dup my mth>opq @
      dup ^class <>
      swap find-opaque @  and  and
      if drop false
      else my mth>xt @ true
      then
    else
      2drop false
    then ;

my macro iv>class " cell+ "
my macro iv>flags " 2 cells+ "
my macro iv>offs " 3 cells+ "
my macro iv>opq " 4 cells+ "
my macro iv>name " 1 chars - dup c@  dup negate chars under+ "

: :ivar ( addr n type-class size -- )
    2>r
    string,|
    ^class my >ifa link,
    2r> over ,
    ^class my >size
      rot 1+ if
        inrec? @ 1+  dup ,
        if cell over +! then
      else 0 ,
      then
    dup @ ,
    +!
    opaque @ , ;

s" SELF" string,|  here
   0 , -1 , 1 , 0 , 0 ,
s" SUPER" string,|  here
over , -1 , 0 , 0 , 0 ,
constant ^super
constant ^self

: mth-exec ( obj mth ) execute ;
: cmth-exec ( obj offs mth ) >r + r>  execute ;
: ivar-exec noop ;
: civar-exec ( obj off ) + ;
: late-exec ( obj mth )
    over cell - @ find-mth
    0= abort" Method not understood by class"
    execute ;

: is-class? ( class -- flag )
    class-link
    begin @ ?dup while
      2dup =
    until 2drop true
    else drop false then ;

: srch-in-class ( str -- str f | xt t )
    ^class dup
    if dup >r my >wl @ >r dup count
      begin 2dup r> search-wordlist  ?dup 0=
      while r> my >sup @ dup
      while dup >r  my >wl @ >r
      repeat
        nip nip
      else r> drop
        2>r 3drop 2r>
      then
    then ;

create mf-buffer 32 chars allot
0 value whats1st
0 value >mfb

: get-mth ( addr n -- mth )
    method-wl search-wordlist 0= abort" method not found" >body ;

: get_ivar/mth ( offs class -- offs' class' [mth] flag )
    mf-buffer count
    2 pick  my >ifa >r
    begin r> @ ?dup while
      dup >r  my iv>name  2over
      find-compare 0=
    until 2drop drop
      r@ my iv>opq @ dup ^class <>
      swap find-opaque @ and  and
      if r> 2drop  false
      else \ ivar gefunden
        r@ my iv>offs @ +
        r> my iv>class @  -1
      then
    else ( offs class addr n )
      method-wl search-wordlist
      if >body 1
      else 2drop false
      then
    then ;

: oopfind ( str -- str f | cfa t )
    srch-in-class ?dup if exit then
    0 to whats1st
    >r  0 to >mfb
    0 0
    begin r@ count >mfb /string  ?dup
    while
      2dup [char] . scan  nip -  ( offs class addr n )
      mf-buffer place \ mf-buffer count type cr
      ?dup
      if get_ivar/mth
      else ^class dup
        if get_ivar/mth dup then
        0=
        if ( obj class off=0 )
          mf-buffer (find) \ suche Klasse/Obj/Mth
          if >body dup is-class? dup to whats1st
            0= if nip  @+
              dup is-class? 0= if 2drop r> false exit then
            then  -1
          else  1 to whats1st
            count method-wl search-wordlist \ Methode (late), dann raus
            if nip >body ['] late-exec  r> drop  true
            else drop r> false
            then exit
          then
        then
      then ( offs class [mth] flag )
      ?dup 0= if r> false exit then
      1+ if swap find-mth
        if whats1st
          if ['] cmth-exec
          else ['] mth-exec
          then r> drop  true
        else drop r> false
        then exit
      then
      mf-buffer c@ +to >mfb
      r@ count >mfb /string
      if c@  [char] . =
        if 1 +to >mfb then
      else drop
      then
    repeat drop
    2dup or
    if r> 2drop  whats1st
      if ['] civar-exec
      else ['] ivar-exec
      then true
    else 2drop  r> false
    then ;

: ivref, ( offs -- )
    postpone op postpone @
    ?dup if postpone literal postpone + then ;

: oopcomp, ( xt -- t | xt f ) \ dup .name .s cr
    case
      ['] mth-exec   of compile, endof
      ['] cmth-exec  of >r postpone literal  postpone +  r> compile, endof
      ['] ivar-exec  of ivref, endof
      ['] civar-exec of ?dup if postpone literal postpone + then endof
      ['] late-exec  of postpone late-exec endof
      false exit
    endcase true ;

: (build) ( class -- )
    create
    dup ,  here swap dup
    my >size @ allot
    ivinit
    does> [ here to obj-xt ] cell+ ;

: (new) ( class -- obj )
    dup my >size @ cell+ allocate abort" Not enough memory"
    2dup !
    cell+ tuck swap ivinit ;

: public: 0 opaque ! ;
: private: 1 opaque ! ;
: protected: ^class opaque ! ;

also forth definitions

: class
    create  here to ^class
    class-link link,
    0 , 0 , ^super , 0 , 0 ,
    public:  save? on  -2 find-opaque !  inrec? off
    ^class             ^self my iv>class  !
    ^class my >sup @  ^super my iv>class  !
    1 #wordlist dup ^class my >wl !
    current @ oldcur !  current !
    also oop
    does> [ here to class-xt ]
      ^class
      if bl word count
        rot dup my >size @
        :ivar
      else (build)
      then ;

: .classes
    class-link
    begin @ ?dup while
      dup body> .name
    repeat ;

:noname ( -<class>- ) ' >body (new) ;
:noname ( -<class>- ) ' >body postpone literal compile (new) ;
interpret/compile: new> immediate

previous definitions

: end-class
    inrec? @ abort" RECORD: closed by END-CLASS"
    previous
    oldcur @ current !
    ^self my iv>class on
    ^super my iv>class on
    -1 find-opaque !
    0 to ^class ;

: super>
    ' >body
    dup ^class my >sup !
           2 cells+
    ^class 2 cells+
    3 cells move ;

: _m;
    _nn;  -cell allot
    save? @
    if s" r> op !" evaluate then
    compile unnest
    mth-xt @ !
    save? on ;

: :m
    method
    ^class my >mfa link,
    ,
    opaque @ ,
    here mth-xt !  0 ,
    :noname drop ['] _m;
    2 here cell- !
    save? @
    if s" op @ >r op !" evaluate then ;

: no-save
    save? off ;

: ns:m
    no-save :m ;

: var
    bl word count
    -1 cell  :ivar ;

: buffer:
    >r
    bl word count
    -1 r>  :ivar ;

: record:
    bl word count
    -1 0 :ivar
    inrec? on ;

: record;
    inrec? off ;

: static ;

find-chain chain-add oopfind
compile,-chain chain-add-before oopcomp,

only forth also definitions

CLASS OBJECT
NS:M CONSTRUCT drop ;
\ STATIC NS:M SIZE my >SIZE @ ;
NS:M PRINT ." Object@" . ;
NS:M DESTRUCT drop ;
END-CLASS

also oop definitions

: _ivinit ( obj class -- )
    dup >r my >ifa
    begin @ ?dup while
      dup my iv>class @ dup 1+
      if ( obj iv ivclass )
        over my iv>offs @  3 pick + swap ( obj iv ivobj ivclass )
        2 pick my iv>flags @ 1 and
        if dup  2 pick cell- ! then
        recurse
      else drop
      then
    repeat
    ( obj )
    [ s" construct" method-wl search-wordlist drop >body ] literal
    r> find-mth if execute else drop then ;

' _ivinit is ivinit

: _ivdest ( obj class -- )
    2dup [ s" destruct" method-wl search-wordlist drop >body ] literal
    swap find-mth if execute else drop then
    my >ifa
    begin @ ?dup while
      dup my iv>class @ dup 1+
      if over my iv>offs @  3 pick + swap  recurse
      else drop
      then
    repeat drop ;

forth definitions oop

: delete ( obj -- )
    dup cell- tuck @ _ivdest
    free abort" failed to delete object" ;

previous definitions

comment:
class listitem super> object
var nextitem
var theitem
:m construct ( item -- ) theitem !  nextitem off ;
end-class

class list super> object
var first
var last
:m construct 0 first !  first self.last ! ;
:m append ( obj -- )
     new> listitem
     dup self.last @ !
     dup listitem.nextitem self.last ! ;
Ns:m print ." List@" . ;
end-class

list mlist
\ comment;

clear-my-macros
.( OK) cr
