\ Assembler

.( Loading Assembler... )

warning off

only forth also definitions

defer setassem

\ vocabulary assembler
assembler vocabulary hidden
only forth also assembler hidden also assembler definitions
also forth also

defer code-@
defer code-!
defer code,
defer code-here

forth ' @ assembler is code-@
forth ' ! assembler is code-!
forth ' h. assembler is code,
forth ' here assembler is code-here

forth

\ *****************
\    The Machine

assembler hidden definitions forth

24 constant framesize

create statframes framesize 8 * allot
0 value <frame>

: framevar
    create ,
    does> @ <frame> + ;

0 framevar <op>
4 framevar <op2>

: <a;>
    <op> @ execute
    <op2> @ <op> !
    ['] noop <op2> ! ;

: <a;!> ( xt -- )
    <op2> ! ;

assembler definitions forth

defer a;
defer a;!

variable postvar

: prefix        ( --- )
    ['] <a;!>    is a;!
    ['] <a;>     is a;
    postvar off ;

: postfix       ( --- )
    ['] execute  is a;!
    ['] noop     is a;
    postvar on ;

prefix          \ default is prefix assembler.

: >pre          2r> postvar @ >r 2>r prefix ;    \ save and set prefix
: pre>          2r> r> if postfix then 2>r ;     \ restore previous fix

\ ***************
\    Modifiers

assembler hidden definitions forth

8 framevar <flags>
   1 constant <upd>
   2 constant <#>
   4 constant <h>
   8 constant <l>
  16 constant <rot>
  32 constant <mask>
  64 constant <mask#>
 512 constant <cy>
 768 constant <ex>
1024 constant <ov>
0x04000000 constant <l/supd>

: mod: ( n -- )
    create ,
    does> @ <flags> @ or <flags> ! ;

assembler definitions forth

    <ov> mod: o
   <upd> mod: .
    <cy> mod: c
    <ex> mod: e
<l/supd> mod: u
     <h> mod: h
     <l> mod: l
   <rot> mod: rot
  <mask> mod: mask
 <mask#> <mask> or
         mod: mask#
     <#> mod: #

<ov> <upd> or mod: o.
<cy> <upd> or mod: c.
<ex> <upd> or mod: e.
<cy> <ov> or mod: co
<ex> <ov> or mod: eo
<cy> <ov> or <upd> or mod: co.
<ex> <ov> or <upd> or mod: eo.

create reg-xts  32 cells allot

: ass-find ( addr n wid -- xt flag | f )
    -rot 2dup 2>r rot
    (voc-search) ?dup
    if 2r> 2drop
    else 2r>
      over c@ 'R' =
      if 1 chars /string
        0 0 2swap >number
        if drop 2drop false
        else 2drop reg-xts +cells true
        then
      else 2drop false
      then
    then ;

' ass-find ' assembler vcfa>voc voc>srchfn !

\ ***************************
\           Registers

assembler hidden definitions forth

16 framevar <regs>

: reg: ( n -<name>- )
    create c,
    does> c@ <regs> count + c!
      <regs> c@ 1+ <regs> c! ;

: rn ( n -- )
    <regs> count + c!
      <regs> c@ 1+ <regs> c! ;
: rn, rn ;

assembler definitions forth
comment:
0 (: dup reg: r| 1+ |
0 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15
16 17 18 19 20 21 22 23 24 25 26 27 28 29 30 31 :)
drop
comment;

0 (: dup reg: r|, 1+ |
0 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15
16 17 18 19 20 21 22 23 24 25 26 27 28 29 30 31 :)
drop

0 (: dup reg: (r|) 1+ |
0 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15
16 17 18 19 20 21 22 23 24 25 26 27 28 29 30 31 :)
drop

comment:
 0 reg: r0    1 reg: r1    2 reg: r2    3 reg: r3
 4 reg: r4    5 reg: r5    6 reg: r6    7 reg: r7    
 8 reg: r8    9 reg: r9   10 reg: r10  11 reg: r11
12 reg: r12  13 reg: r13  14 reg: r14  15 reg: r15
16 reg: r16  17 reg: r17  18 reg: r18  19 reg: r19
20 reg: r20  21 reg: r21  22 reg: r22  23 reg: r23
24 reg: r24  25 reg: r25  26 reg: r26  27 reg: r27
28 reg: r28  29 reg: r29  30 reg: r30  31 reg: r31

 0 reg: r0,    1 reg: r1,    2 reg: r2,    3 reg: r3,
 4 reg: r4,    5 reg: r5,    6 reg: r6,    7 reg: r7,    
 8 reg: r8,    9 reg: r9,   10 reg: r10,  11 reg: r11,
12 reg: r12,  13 reg: r13,  14 reg: r14,  15 reg: r15,
16 reg: r16,  17 reg: r17,  18 reg: r18,  19 reg: r19,
20 reg: r20,  21 reg: r21,  22 reg: r22,  23 reg: r23,
24 reg: r24,  25 reg: r25,  26 reg: r26,  27 reg: r27,
28 reg: r28,  29 reg: r29,  30 reg: r30,  31 reg: r31,

 0 reg: (r0)    1 reg: (r1)    2 reg: (r2)    3 reg: (r3)
 4 reg: (r4)    5 reg: (r5)    6 reg: (r6)    7 reg: (r7)    
 8 reg: (r8)    9 reg: (r9)   10 reg: (r10)  11 reg: (r11)
12 reg: (r12)  13 reg: (r13)  14 reg: (r14)  15 reg: (r15)
16 reg: (r16)  17 reg: (r17)  18 reg: (r18)  19 reg: (r19)
20 reg: (r20)  21 reg: (r21)  22 reg: (r22)  23 reg: (r23)
24 reg: (r24)  25 reg: (r25)  26 reg: (r26)  27 reg: (r27)
28 reg: (r28)  29 reg: (r29)  30 reg: (r30)  31 reg: (r31)
comment;

31 reg: ip    31 reg: ip,    31 reg: (ip)
30 reg: rp    30 reg: rp,    30 reg: (rp)
29 reg: sp    29 reg: sp,    29 reg: (sp)
28 reg: tos   28 reg: tos,   28 reg: (tos)
27 reg: op    27 reg: op,    27 reg: (op)
26 reg: lp    26 reg: lp,    26 reg: (lp)
25 reg: data0 25 reg: data0, 25 reg: (data0)
 1 reg: ssp    1 reg: ssp,    1 reg: (ssp)
 2 reg: toc    2 reg: toc,    2 reg: (toc)
 3 reg: cfa    3 reg: cfa,    3 reg: (cfa)
12 reg: rtemp 12 reg: rtemp, 12 reg: (rtemp)

65 reg: xer 65 reg: xer,
72 reg: lr  72 reg: lr,
73 reg: ctr 73 reg: ctr,

32 reg: cr0   33 reg: cr1    34 reg: cr2
35 reg: cr3   36 reg: cr4    37 reg: cr5
38 reg: cr6   39 reg: cr7
32 reg: cr0,  33 reg: cr1,   34 reg: cr2,
35 reg: cr3,  36 reg: cr4,   37 reg: cr5,
38 reg: cr6,  39 reg: cr7,

assembler hidden definitions forth

: freg: ( n -<name>- )
    128 + reg: ;

: fn ( n -- )
    128 + rn ;

assembler definitions forth

 0 freg: f0    1 freg: f1    2 freg: f2    3 freg: f3
 4 freg: f4    5 freg: f5    6 freg: f6    7 freg: f7    
 8 freg: f8    9 freg: f9   10 freg: f10  11 freg: f11
12 freg: f12  13 freg: f13  14 freg: f14  15 freg: f15
16 freg: f16  17 freg: f17  18 freg: f18  19 freg: f19
20 freg: f20  21 freg: f21  22 freg: f22  23 freg: f23
24 freg: f24  25 freg: f25  26 freg: f26  27 freg: f27
28 freg: f28  29 freg: f29  30 freg: f30  31 freg: f31

 0 freg: f0,    1 freg: f1,    2 freg: f2,    3 freg: f3,
 4 freg: f4,    5 freg: f5,    6 freg: f6,    7 freg: f7,    
 8 freg: f8,    9 freg: f9,   10 freg: f10,  11 freg: f11,
12 freg: f12,  13 freg: f13,  14 freg: f14,  15 freg: f15,
16 freg: f16,  17 freg: f17,  18 freg: f18,  19 freg: f19,
20 freg: f20,  21 freg: f21,  22 freg: f22,  23 freg: f23,
24 freg: f24,  25 freg: f25,  26 freg: f26,  27 freg: f27,
28 freg: f28,  29 freg: f29,  30 freg: f30,  31 freg: f31,

assembler hidden definitions forth

12 framevar <cond>

: cond: ( n -- )
    create dup 0>=
      if 16 lshift
      else drop 0 then
      swap 0x1f and 21 lshift or ,
    does> @ <cond> @ or <cond> ! ;

assembler definitions forth

16 -1 cond: dnz
18 -1 cond: dz
8 0 cond: lt      4 0 cond: ge
8 1 cond: gt      4 1 cond: le
8 2 cond: eq      4 2 cond: ne
8 3 cond: so      4 3 cond: ns

\ ***********
\    Reset

: reset ( -- )
    <frame> framesize 2 cells /string erase ['] noop <op> ! ;

\ *************************
\    Instruction Formats

assembler hidden definitions forth

: I, ( lit opcode -- )
    swap 0x3fffffc and or code, ;

: B, ( bo bi disp opcode -- )
    swap 0xfffc and or -rot
    ( 0x1f and ) 16 lshift swap
    ( 0x1f and ) 21 lshift or or
    code, ;

: SC, ( opcode -- )
    code, ;

: D, ( rt ra nr16 opcode -- )
    or -rot
    ( 0x1f and ) 16 lshift swap
    ( 0x1f and ) 21 lshift or or
    code, ;

: X, ( ra rs rb opcode -- )
    >r
    ( 0x1f and ) 11 lshift swap
    ( 0x1f and ) 21 lshift or swap
    ( 0x1f and ) 16 lshift or
    r> or
    code, ;

: XFX, ( rt spr opcode -- )
    -rot
    ( 0x1f and ) 11 lshift swap
    ( 0x1f and ) 21 lshift or or
    code, ;
    
: XFL, ( flm frb opcode -- )
    -rot
    ( 0x3ff and ) 11 lshift swap
    ( 0xff and ) 17 lshift or or
    code, ;
    
: XL, ( bo bi opcode -- )
    -rot
    ( 0x1f and ) 16 lshift swap
    ( 0x1f and ) 21 lshift or or
    code, ;
    
: XO, ( rt ra rb opcode -- )
    >r
    ( 0x1f and ) 11 lshift swap
    ( 0x1f and ) 16 lshift or swap
    ( 0x1f and ) 21 lshift or
    r> or
    code, ;

: M, ( ra rs sh mb me opcode -- )
    >r
    ( 0x1f and ) 1 lshift swap
    6 lshift or swap
    11 lshift or swap
    21 lshift or swap
    16 lshift or
    r> or
    code, ;

\ *****************************
\    Argument classification

: ?r ( r -- r t | f )
    dup 0xe0 and
    if drop false
    else true
    then ;

: ?f ( r -- r t | f )
    dup 0xe0 and 0x80 =
    if 0x1f and true else drop false then ;

: ?rrr ( -- r1 r2 r3 t | f )
    <regs> count
    3 = if
      count ?r >r swap
      count ?r >r swap
      c@    ?r
      r> and r> and
      if true
      else 2drop drop false
      then
    else
      drop false
    then ;

: ?rr# ( -- r1 r2 t | f )
    <regs> count
    2 =  <flags> @ <#> and and
    if count ?r >r swap c@ ?r
      r> and
      if <flags> dup @ <#> xor swap !
        true
      else 2drop false
      then
    else
      drop false
    then ;

: ?rr ( -- r1 r2 t | f )
    <regs> count
    2 =
    if count ?r >r swap c@ ?r
      r> and
      if true
      else 2drop false
      then
    else
      drop false
    then ;

: ?fff ( -- r1 r2 r3 t | f )
    <regs> count
    3 = if
      count ?f >r swap
      count ?f >r swap
      c@    ?f
      r> and r> and
      if true
      else 2drop drop false
      then
    else
      drop false
    then ;

: ?ff# ( -- r1 r2 t | f )
    <regs> count
    2 =  <flags> @ <#> and and
    if count ?f >r swap c@ ?f
      r> and
      if <flags> dup @ <#> xor swap !
        true
      else 2drop false
      then
    else
      drop false
    then ;


: ?. ( opcode -- opcode' )
    <flags> @ <upd> and
    or ;

: ?o. ( opcode -- opcode' )
    <flags> @ [ <ov> <upd> or ] literal and
    or ;

\ ****************
\   Instructions

: (op:) ( -- ad )
    current @
    also assembler definitions previous
    create here 0 ,
    does> @ a;! a; ;

: op:
    (op:) :noname ;

: ;op
    postpone ;
    swap !
    current ! ; immediate

: >s# ( n -- n' )
    dup 0xffff and
    if dup 0x8000 + 16 rshift
      abort" Illegal signed number"
      0xffff and
    else 16 rshift
      0x4000000 or
    then ;

: (mov#) ( rt # -- )
    over >r dup 0xffff and ?dup over 0= or
    if r@ 0 rot 0x38000000 d, r>
    else r>drop 0 then swap
    0x8000 + 16 rshift ?dup
    if 0x3c000000 d, else 2drop then ;

op: mov ( -- )
    <regs> count
    dup 1 =
    if drop <flags> @ <#> and 0= abort" must mov with #"
      c@ ?r 0= abort" must mov to a reg"
      swap (mov#)
    else 2 <> abort" must mov with reg/#"
      count dup 0x40 and
      if 0x1f and 5 lshift
        swap c@ ?r 0= abort" illegal register" \ mtspr
        swap 0x7c0003a6 XFX,
      else swap c@ dup 0x40 and
        if 0x1f and 5 lshift
          0x7c0002a6 XFX, \ mfspr
        else dup 0x7c000378 X, \ or
        then
      then
    then
    reset ;op

: (b)
    <cond> @ 0x03e00000 and 0<> negate
    <regs> count bounds
    ?do i c@ dup 0x20 and
      if 7 and 18 lshift  <cond> @ 0xffe3ffff and  or  <cond> !
        dup 0= -
      else dup 0x40 and
        if nip 70 -
        else true abort" branch to lr/ctr"
        then
      then
    loop
    ?dup if >r
      <cond> @
      dup 21 rshift 0x1f and
      dup 0xc and
      dup 0x8 <>
      if dup 0x4 =
        if drop 0
        else 0<> abort" Illegal condition"
          16
        then
      then
      swap dup 0x10 and
      if 2 and or
      else drop 4 or
      then
      swap 16 rshift 0x1f and
      <cond> @ 0x200001 and  r>
      1- ?dup if 1- if 0x4c000420 else 0x4c000020 then
        or XL,
      else 0x40000000 or
        3 pick code-here - swap B, drop
      then
    else code-here - 0x48000000 <cond> @ or I,
    then reset ;

op: b   (b) ;op
op: bl  <cond> incr (b) ;op
op: bc  code-here - 0x40000000 B, reset ;op
op: bcl code-here - 0x40000001 B, reset ;op

: (ld/st) ( op -- )
    >r <regs> count
    2 =
    if count swap c@ rot r> <flags> @ or swap 0xffff and D,
    else count swap count -rot c@
      0x7c00002e
      r@ <flags> @ or
         0x6c000000 and 20 rshift or
      r> 0x10000000 and 16 rshift or X,
    then
    reset ;

op: lbz 0x88000000 (ld/st) ;op
op: lhz 0xa0000000 (ld/st) ;op
op: lha 0xa8000000 (ld/st) ;op
op: lwz 0x80000000 (ld/st) ;op

op: stb 0x98000000 (ld/st) ;op
op: sth 0xd0000000 (ld/st) ;op
op: stw 0x90000000 (ld/st) ;op

op: cntlz ( -- )
  ?rr if
    0 0x7c000034 ?. X,
  else
    true abort" Illegal CNTLZ"
  then
  reset ;op

: (ext) ( opcode -- )
  ?rr if
    rot 0 swap ?. X,
  else
    true abort" Illegal EXT"
  then
  reset ;

op: extsb ( -- ) 0x7c000778 (ext) ;op
op: extsh ( -- ) 0x7c000734 (ext) ;op

op: neg ( -- )
  ?rr if
    0 0x7c0000d0 ?o. XO,
  else true abort" Illegal NEG" then  
  reset ;op

op: ins ( -- )
  ?rr <flags> @ [ <rot> <#> or ] literal and
  [ <rot> <#> or ] literal = and if
    rot >r 2swap r>
    0x50000000 ?. M,
  else true abort" Illegal INS" then
  reset ;op

op: sra ( -- )
  ?rr# if
    rot 0x7c000670 ?. X,
  else
    ?rrr 0= abort" Illegal SRA"
    0x7c000630 ?. X,
  then
  reset ;op

op: shl ( -- )
  ?rr# if
    rot 0 31 2 pick - 0x54000000 ?. M,
  else
    ?rrr 0= abort" Illegal SHL"
    0x7c000030 ?. X,
  then
  reset ;op

op: shr ( -- )
  ?rr# if
    rot 32 over - swap 31 0x54000000 ?. M,
  else
    ?rrr 0= abort" Illegal SHR"
    0x7c000430 ?. X,
  then
  reset ;op

op: eqv ( -- )
  ?rrr 0= abort" Illegal EQV"
  0x7c000238 ?. X,
  reset ;op

op: nor ( -- )
  ?rrr 0= abort" Illegal NOR"
  0x7c0000f8 ?. X,
  reset ;op

op: nand ( -- )
  ?rrr 0= abort" Illegal NAND"
  0x7c0003b8 ?. X,
  reset ;op

op: xor ( -- )
  ?rr# if
    rot dup 0xffff and ?dup
    if swap 3 pick 2>r
      0x68000000 D,
      2r> dup rot 16 rshift ?dup
      if 0x6c000000 D,
      else 2drop
      then
    else 16 rshift 0x6c000000 D,
    then
  else
    ?rrr 0= abort" Illegal XOR"
    0x7c000278 ?. X,
  then
  reset ;op

op: or ( -- )
  <flags> @ <cy> and if
    ?rrr 0= abort" Illegal OR"
    0x7c000338 ?. X,
  else
    ?rr# if
      rot dup 0xffff and ?dup
      if swap 3 pick 2>r
        0x60000000 D,
        2r> dup rot 16 rshift ?dup
        if 0x64000000 D,
        else 2drop
        then
      else 16 rshift 0x64000000 D,
      then
    else
      ?rrr 0= abort" Illegal OR"
      0x7c000378 ?. X,
    then
  then
  reset ;op

create logtab
(: 31 | - c, | 0 1 2 6 3 11 7 16 4 14 12 21 8 23 17 26 31 5 10 15 13 20 22 25 30 9 19 24 29 18 28 27 :)

op: and ( -- )
  <flags> @ <cy> and if
    ?rrr 0= abort" Illegal AND"
    0x7c000078 ?. X,
  else
    <flags> @ <rot> and if
      <flags> @ dup <mask> and 0= abort" AND ROT requires MASK n1 n2 or MASK# n"
      <mask#> and
      if dup dup dup 1- and xor dup under+ swap
        dup dup 1- and abort" Invalid MASK#"
        0x4653adf * 27 rshift 0x1f and logtab + c@ 1+ swap
        dup dup 1- and abort" Invalid MASK#"
        0x4653adf * 27 rshift 0x1f and logtab + c@
      then
      ?rr# if
        rot >r 2swap r> 0x54000000 ?. M,
      else ?rrr 0= abort" Illegal AND ROT"
        >r 2swap r> -rot 0x5c000000 ?. M,
      then
    else
      ?rr# if
        <flags> @ <upd> and 0= abort" AND # sets CR0"
        swap rot dup 0xffff and ?dup
        if swap 16 rshift abort" Illegal AND #"
          0x70000000 D,
        else 16 rshift 0x74000000 D,
        then
      else
        ?rrr 0= abort" Illegal AND"
        0x7c000038 ?. X,
      then
    then
  then
  reset ;op

op: cmp
  <flags> @ <#> and if
    <regs> count 1 =
    if 0 else count 7 and 2 lshift then
    swap c@ swap rot 0xffff and 0x2c000000
    <flags> @ <l/supd> and xor D,
  else
    ?rr if 0 swap
    else <regs> 1+ count 7 and 2 lshift swap count -rot c@
    then
    0x7c000000
    <flags> @ <l/supd> and 20 rshift or X,
  then
  reset ;op

op: div
  ?rrr 0= abort" Illegal DIV"
  0x7c0003d6 ?o.
  <flags> @ <l/supd> and 20 rshift xor XO,
  reset ;op

op: mul
  <flags> @ <h> and if
    ?rrr 0= abort" Illegal MUL H"
    0x7c000096
    <flags> @ <l/supd> and 19 rshift xor ?. XO,
  else
    <flags> @ <#> if
      ?rr 0= abort" Illegal MUL #"
      rot 0xffff and
      0x1c000000 D,
    else
      ?rrr 0= abort" Illegal MUL L"
      0x7c0001d6 ?o. XO,
    then
  then
  reset ;op

: _(add#) ( nr dreg sreg opcode -- )
    >r rot <flags> @ 0=
    if r@ 4 and
      if r>drop
        dup 0xffff and ?dup
        if 2over rot 0x38000000 D,  >r drop dup r>
        then
        0x8000 + 16 rshift ?dup
        if 0x3c000000 d, else 2drop then
        exit
      then
    then
    <flags> @ <ex> and
    if dup 0 = over -1 = or
      if 0x40 and r> 0x1e and 0x7c000180 or or
        <flags> @ [ <ov> <upd>  or ] literal and or
        XO,
        exit
      then
    then
    <flags> @ <cy> and
    if dup 0x8000 + 16 rshift 0=
      if 0xffff and  r> 4 and 26 rshift 0x20000000 or
        dup 0x20000000 > if <flags> @ <upd> and 26 rshift or
        else true abort" Illegal SUBF C. #" then
        D,
        exit
      then
    then
    rtemp swap (mov#)  rtemp
    <flags> @ dup [ <cy> <ex> or ] literal and
    if dup <ex> and
      r> dup 4 and 0= if 224 + then
      xor  swap
    then
    [ <ov> <upd> or ] literal and or
    XO, ;

: (add#) 0x7c000214 _(add#) ;

: (rsb#) 0x7c000050 _(add#) ;

: (addreg) ( dreg sreg1 sreg2 -- )
    0x7c000214
    <flags> @ 0x701 and xor
    XO, ;

: (rsbreg) ( dreg sreg1 sreg2 -- )
    0x7c000050
    <flags> @ 0x401 and or
    <flags> @ 0x200 and 3 rshift xor
    <flags> @ 0x100 and 3 * 2 rshift +
    XO, ;

op: add
   ?rr# if (add#) else ?rrr if (addreg) else true abort" Unknown ADD format" then then
   reset ;op

op: subf
   ?rr# if (rsb#) else ?rrr if (rsbreg) else true abort" Unknown SUBF format" then then
   reset ;op

op: sub
   ?rr# if rot negate -rot (add#)
   else ?rrr
     if swap (rsbreg)
     else true abort" Unknown SUB format"
     then
   then
   reset ;op

\ *****************************
\        Local labels

 32 value max-llabs
 12 value b/llab
false value ll-global?     \ are local labels available globally?

create %llab[] max-llabs b/llab * allot

%llab[] value llab[]            \ default to %llab[] array

false value ll-used?

: llab-init  ( -- )     \ initializes local labels
  llab[]  max-llabs b/llab * erase
  false to ll-used? ;

: llab>line  ( n -- ^line )
  dup max-llabs 1- u> abort" Bad Label"
  b/llab * llab[] + ;

: >res  ( ^line -- )
  cell+ @ dup 0= IF \ if nothing to resolve
    drop exit       \   then exit
  THEN
  BEGIN             \ stack contains directory address of 
                    \   displacement to be resolved
    dup code-@ >r 
    dup code-here -
    r@ 0xfc000000 and 0x48000000 =
    if 0x03fffffc and r@ 0xfc000003 and or
      r> dup 6 lshift 0<
      if 0xfc000000 or -4 and
      else 0x3fffffc and then
    else 0xfffc and r@ 0xffff0003 and or
      r> dup 16 lshift 0<
      if 0xffff0000 or -4 and
      else 0xfffc and then 
    then >r
    over code-!
    r> ?dup
  WHILE
    +
  REPEAT drop ;

: $:f  ( n -- )     \ defines a local label
  true to ll-used?  \ set "labels used?" flag
  llab>line
  dup @ 0<> abort" Label can't be multiply defined"
  dup >res          \ resolve forward references if needed
  here swap ! ;     \ and set label for subsequent refs

assembler definitions

: $  ( n1 -- n2 )
  true to ll-used?          \ set "labels used?" flag
  llab>line 1 over 8 + c!   \ set "ever referenced?" flag
  dup @ IF      \ if the label is already defined:
    @           \   then return it for resolution
  ELSE          \ otherwise:
    cell+       \   move to head of list pointer
    dup @ >r    \   save old head of list on rstack
    here swap ! \   set new head of list
    r>          \   retrieve old head of list
    dup 0= IF   \   if list is empty:
      code-here +    \     pass current dictionary location
    THEN        \   end-if
  THEN ;        \ end-if

: $:  ( n -- )      \ allow use as prefix/postfix
  ['] $:f a;! a; ;

hidden definitions

: _ll-errs?  ( -- )  \ final error checking for local labels
  false max-llabs 0 DO  \ check each label
    i b/llab * llab[] +
    dup 8 + c@ 0<> IF   \ if jumps to label
      @ 0= IF           \   and no label to jump to
        cr ." jump(s) to label " i .
          ." and label not defined"
        drop true       \ set error flag
      THEN
    ELSE                \ if no jumps to label
      @ 0<> IF          \   and label defined
        cr ." warning - label " i .
           ." defined, but no jumps to it"
      THEN
    THEN
  LOOP
  IF abort THEN ;       \ abort if fatal error

: ll-errs?  ( -- )      \ final error checking for local labels
  ll-used? IF _ll-errs? THEN ;

FORTH DEFINITIONS

: DOASSEM       ( --- ) 
    ['] DROP A;!
    <op2> 2@ <op> 2!
    statframes to <frame>  reset
    ll-global? 0= if llab-init then               \ in case labels used
    ALSO assembler ;

' DOASSEM IS SETASSEM

synonym CLEAR_LABELS LLAB-INIT

: LOCAL_REF     ( --- )
                0 to LL-GLOBAL? ;       LOCAL_REF
                                        \ default to LOCAL references only

: GLOBAL_REF    ( --- )
                -1 to LL-GLOBAL? ;

: LABEL         ( NAME --- )            \ Really just a constant addr
                SETASSEM CREATE ; 

: CODE          ( NAME --- )
                LABEL HIDE ;

assembler hidden DEFINITIONS forth

\ **********************************
\    Forth conditional structures

: a>resolve
    dup code-@ dup>r 26 rshift 16 =
    if dup code-here - 0xfffc and r> 0xffff0003 and
    else r>drop dup code-here - 0x3fffffc and 0x48000000
    then or swap code-! ;

assembler definitions forth

0x00800000 constant 0<   0x01800000 constant 0>=
0x00810000 constant 0>   0x01810000 constant 0<=
0x00820000 constant 0=   0x01820000 constant 0<>

: if     >r a; code-here 0 0 0 0x40000000 r> or I, ;
: else   a; >r code-here
    0 0x48000000 I,
    r> a>resolve ;
: then   a; a>resolve ;

: begin  a; code-here ;
: again  a; code-here - 0x48000000 I, ;
: until  >r a; 0 0 rot code-here - 0x40000000 r> or B, ;
: while  >r a; code-here 0 0 0 0x40000000 r> or B, swap ;
: repeat [ also assembler ] again [ previous ] a>resolve ;

\ **********************
\    Assembler Macros

assembler hidden definitions forth

: mreg: ( nr -- )
    create c,
    does> c@ <regs> framesize - + c@ rn, ;

assembler definitions forth

1 mreg: M1   1 mreg: M1,
2 mreg: M2   2 mreg: M2,
3 mreg: M3   3 mreg: M3,
4 mreg: M4   4 mreg: M4,

assembler hidden definitions forth

: <mflags>
    <flags> framesize - ;
: enter-macro
    framesize +to <frame> reset ;

: leave-macro
    a;
    framesize negate +to <frame>
    reset ;

: (mac:)
    create here 0 ,
    does> @ a;! a; ;

assembler definitions forth

: copyflags
    <flags> dup framesize - @ swap ! ;

: mac:
    current @ >r
    also assembler definitions
    (mac:)
    :noname
    postpone enter-macro
    r> current ! ;

: ;mac
    previous
    postpone leave-macro
    postpone ;
    swap ! ; immediate

mac: next
  b copyflags lr
;mac

mac: li
  add m1, r0, #
;mac

mac: nop
  or r0, r0, r0
;mac

mac: not
  nor m1, m2, m2
;mac

mac: rol
  [ forth ] <mflags> @ <#> and if [ assembler ]
    and m1, m2 rot # mask# -1
  [ forth ] else [ assembler ]
    and m1, m2, rot m3 mask# -1
  [ forth ] then [ assembler ]
;mac

mac: ror
  [ forth ] <mflags> @ <#> and if [ assembler ]
    and m1, m2 rot # 32 swap - mask# -1
  [ forth ] else true abort" Illegal ROR" then [ assembler ]
;mac

: END-CODE
    ll-global? 0=
    if ll-errs? then        \ check for local label errors
    PREVIOUS A; REVEAL ;

synonym C; END-CODE

only forth also definitions
warning on
.( OK) cr