// Forth machine
#include <stdio.h>
#include <stdlib.h>
#include <image.h>
#include <unistd.h>
#include <termios.h>

#define TCELL unsigned long
#define CELL sizeof(TCELL)
#define CFASIZE 2*CELL

#define CMEM(ad) *((TCELL *)(memory+ad))
#define CHMEM(ad) memory[ad]
#define WMEM(ad) *((unsigned short *)(memory+ad))

#define ROUTINES(ad) **(void **)(rtns+ad)
#define PUSH *(--sp)=tos;
#define POP tos=*sp++;

#define MYUP(c) (((c>96)&&(c<123))?c-32:c)

char pbuf[256];
int cnt;

typedef (*libcall00)();
typedef int (*libcall0i)();
typedef unsigned char *(*libcall0p)();
typedef long long(*libcall0d)();
typedef (*libcalli0)(int);
typedef int (*libcallii)(int);
typedef unsigned char *(*libcallip)(int);
typedef long long(*libcallid)(int);
typedef (*libcallp0)(unsigned char *);
typedef int (*libcallpi)(unsigned char *);
typedef unsigned char *(*libcallpp)(unsigned char *);
typedef long long(*libcallpd)(unsigned char *);
typedef (*libcalld0)(long long);
typedef int (*libcalldi)(long long);
typedef unsigned char *(*libcalldp)(long long);
typedef long long(*libcalldd)(long long);

//typedef (*libcall00)();
typedef int (*libcall01)();
typedef long long (*libcall02)();
typedef (*libcall10)(int);
typedef int (*libcall11)(int);
typedef long long (*libcall12)(int);
typedef (*libcall20)(int,int);
typedef int (*libcall21)(int,int);
typedef long long (*libcall22)(int,int);
typedef (*libcall30)(int,int,int);
typedef int (*libcall31)(int,int,int);
typedef long long (*libcall32)(int,int,int);
typedef (*libcall40)(int,int,int,int);
typedef int (*libcall41)(int,int,int,int);
typedef long long (*libcall42)(int,int,int,int);
typedef (*libcall50)(int,int,int,int,int);
typedef int (*libcall51)(int,int,int,int,int);
typedef long long (*libcall52)(int,int,int,int,int);
typedef (*libcall60)(int,int,int,int,int,int);
typedef int (*libcall61)(int,int,int,int,int,int);
typedef long long (*libcall62)(int,int,int,int,int,int);
typedef (*libcall70)(int,int,int,int,int,int,int);
typedef int (*libcall71)(int,int,int,int,int,int,int);
typedef long long (*libcall72)(int,int,int,int,int,int,int);
typedef (*libcall80)(int,int,int,int,int,int,int,int);
typedef int (*libcall81)(int,int,int,int,int,int,int,int);
typedef long long (*libcall82)(int,int,int,int,int,int,int,int);
typedef (*libcall90)(int,int,int,int,int,int,int,int,int);
typedef int (*libcall91)(int,int,int,int,int,int,int,int,int);
typedef long long (*libcall92)(int,int,int,int,int,int,int,int,int);

#define PRINTNAME(ad) if(CMEM(0)==0) { memset(pbuf,' ',cnt); memcpy(pbuf+cnt,memory+(ad-9-(CHMEM(ad-9)&0x3f)),(CHMEM(ad-9)&0x3f)); \
pbuf[(CHMEM(ad-9)&0x3f)+cnt]=0; printf("%s ",pbuf); for(tsp=spn;tsp>=sp;tsp--) printf("%x ",*tsp); printf("%x\n",tos);}
//#define NEXT for(tsp=spn;tsp>=sp;tsp--) printf("%x ",*tsp); printf("%x ",tos); printf("\n%x ",cfa=CMEM(ip)); ip+=CELL; goto ROUTINES(CMEM(cfa));
//#define NEXT cfa=CMEM(ip); PRINTNAME(cfa); ip+=CELL; goto ROUTINES(CMEM(cfa));
#define NEXT cfa=CMEM(ip); ip+=CELL; goto ROUTINES(CMEM(cfa));

char tbuf[256];

machine(unsigned char *memory)
{ TCELL *sp,*rp,ip,cfa,tos,help/* ,*tsp,*spn */ ;
  void *rtns;
  void *lp;

  static void *routines[] = {
    &&mcrash,
    &&xtprim,
    &&xtcol,
    &&xtunnest,
    &&xtvar,
    &&xtcon,
    &&xtdef,
    &&xtdoes,
    &&xtval,
    &&xtvalst,
    &&xtvalpst,
    &&xtlocfetch,
    &&xtlocstore,
    &&xtlocpstore,
    
    &&xtcall00,
    &&xtcall0i,
    &&xtcall0p,
    &&xtcall02,
    &&xtcalli0,
    &&xtcallii,
    &&xtcallip,
    &&xtcalli2,
    &&xtcallp0,
    &&xtcallpi,
    &&xtcallpp,
    &&xtcallp2,
    &&xtcall20,
    &&xtcall2i,
    &&xtcall2p,
    &&xtcall22,
    &&xtcall,

    &&xtlocinit,
    &&xtunparm,
    &&xtlocalloc,

    &&xtlit,
    &&xtexec,
    &&xtperf,
    &&xtnoop,

    &&xtunnest,
    &&xtbra,
    &&xtqbra,
    
    &&xti,
    &&xtj,
    &&xtk,

    &&xtdrop,
    &&xtdup,
    &&xtswap,
    &&xtover,
    &&xtrot,
    &&xtmrot,
    &&xttuck,
    &&xtpick,
    &&xtnip,
    &&xtqdup,
    &&xttdrop,
    &&xttdup,
    &&xttswap,
    &&xttover,
    &&xttrot,
    &&xtthdrop,
    &&xtfdrop,
    &&xtspfetch,
    &&xtspstore,

    &&xtplus,
    &&xtminus,
    &&xtunderp,
    &&xtnegate,
    &&xtabs,
    &&xttimes,
    &&xtumt,
    &&xtmt,
    &&xtdiv,
    &&xtumdmod,
    &&xtsmdrem,
    &&xtfmdmod,
    &&xtdmod,
    &&xttdmod,

    &&xtdpl,
    &&xtdmi,
    &&xtdneg,
    &&xtdabs,
    &&xtstd,

    &&xttor,
    &&xtfromr,
    &&xtrfetch,
    &&xtttor,
    &&xttfromr,
    &&xttrfetch,
    &&xtrpfetch,
    &&xtrpstore,
    &&xtduptor,
    &&xtrdrop,

    &&xtstore,
    &&xtfetch,
    &&xtpstore,
    &&xtcstore,
    &&xtcfetch,
    &&xtcpstore,
    &&xtwstore,
    &&xtwfetch,
    &&xtwpstore,
    &&xttstore,
    &&xttfetch,
    &&xtlcnt,
    &&xtpstoreplus,
    &&xtwrdspl,
    &&xtwrdjoin,

    &&xtcnt,
    &&xtwcnt,
    &&xtlcnt,
    &&xtzcnt,

    &&xtopl,
    &&xtomi,
    &&xtcpl,
    &&xtcmi,
    &&xtcs,
    &&xtcspl,
    &&xtcsmi,
    &&xtplcs,
    &&xtmics,
    &&xtttimes,
    &&xttdiv,
    &&xtutdiv,
    &&xtchspl,
    &&xtdttim,
    &&xtdtdiv,

    &&xtnoop,
    &&xtqbra,
    &&xtqbra,
    &&xtbra,
    &&xtbra,
    &&xtqbra,
    &&xtbra,
    &&xtnoop,
    &&xtdo,
    &&xtqdo,
    &&xtloop,
    &&xtploop,
    &&xtnoop,
    &&xtof,
    &&xtbra,
    &&xtdrop,
    &&xtbra,
    &&xtunloop,

    &&xtmove,
    &&xtcmove,
    &&xtcmove2,
    &&xtfill,

    &&xtand,
    &&xtor,
    &&xtxor,
    &&xtinv,
    &&xtlsh,
    &&xtrsh,

    &&xtincr,
    &&xtdecr,
    &&xtcincr,
    &&xtcdecr,
    &&xttoggle,

    &&xtzeq,
    &&xtzneq,
    &&xtzlss,
    &&xtzgrt,
    &&xteq,
    &&xtneq,
    &&xtlss,
    &&xtgrt,
    &&xteq,
    &&xtneq,
    &&xtulss,
    &&xtugrt,
    &&xtdulss,
    &&xtbetween,
    &&xtwithin,

    &&xttrailch,
    &&xtcutstr,

    &&xtuemit,
    &&xtutype,
    &&xtubye,
    &&xtucr,
    &&xtugotoxy,
    &&xtukey,
    &&xtukeyq,

    &&xtopfi,
    &&xtcrfi,
    &&xtclfi,
    &&xtrdfi,
    &&xtwrfi,
    &&xtposfi,
    &&xtrpfi,
    &&xtfipos,

    &&xtprswht,
    &&xtcomp,
    &&xtccomp,
    &&xthash,

    &&xtallocate,
    &&xtresize,
    &&xtfree,
    
    &&xtnextimage,
    &&xtimgsymb,
    &&xtloadaddon,
    &&xtabs2rel,
    &&xtrel2abs,
    0
};
static const void *mcalls[] = {
&&mc00,&&mc10,&&mc20,&&mc30,&&mc40,&&mc50,&&mc60,&&mc70,&&mc80,&&mc90,
&&mc01,&&mc11,&&mc21,&&mc31,&&mc41,&&mc51,&&mc61,&&mc71,&&mc81,&&mc91,
&&mc02,&&mc12,&&mc22,&&mc32,&&mc42,&&mc52,&&mc62,&&mc72,&&mc82,&&mc92};

//  printf("Memory is at %x\n",memory);
//cnt=0; memset(pbuf,' ',16);
rp=(TCELL *)(memory+CMEM(2*CELL));
//  printf("Return Stack is %x\n",rp);
sp=(TCELL *)(rp-CMEM(4*CELL));
//spn=sp; CMEM(0)=0;
//  printf("Data Stack is %x\n",sp);
  rtns=(char *)routines; lp=0;
  POP;
  ip=CELL;
//  printf("Go!\n");
  NEXT;
mcrash:
  exit(1);
xtprim:
  goto ROUTINES(CMEM(cfa+CELL));
xtcol:
  *--rp=ip;
  ip=cfa+CFASIZE;
//  cnt++;
  NEXT;
xtunnest:
  ip=*rp++; //cnt--;
  NEXT;
xtvar:
  PUSH;
  tos=cfa+CFASIZE;
  NEXT;
xtcon:
  PUSH;
  tos=CMEM(cfa+CFASIZE);
  NEXT;
xtdef:
  cfa=CMEM(cfa+CFASIZE); goto ROUTINES(CMEM(cfa));
  NEXT;
xtdoes:
  PUSH;
  *--rp=ip;
  ip=CMEM(cfa+CELL);
  tos=cfa+CFASIZE;
//  cnt++;
  NEXT;
xtval:
  PUSH;
  tos=CMEM(cfa+CFASIZE);
  NEXT;
xtvalst:
  CMEM(cfa-CELL)=tos;
  POP;
  NEXT;
xtvalpst:
  CMEM(cfa-2*CELL)+=tos;
  POP;
  NEXT;
xtlocfetch:
  PUSH;
  tos=*(TCELL *)(lp+CMEM(cfa+CFASIZE));
  NEXT;
xtlocstore:
  *(TCELL *)(lp+CMEM(cfa-CELL))=tos;
  POP;
  NEXT;
xtlocpstore:
  *(TCELL *)(lp+CMEM(cfa-2*CELL))+=tos;
  POP;
  NEXT;
xtcall00:
  (*(libcall00)(CMEM(cfa+CELL)))();
  NEXT;
xtcall0i:
  PUSH;
  tos=(*(libcall0i)(CMEM(cfa+CELL)))();
  NEXT;
xtcall0p:
  PUSH;
  tos=(*(libcall0p)(CMEM(cfa+CELL)))()-memory;
  NEXT;
xtcall02: { long long hlp;
    PUSH;
    hlp=(*(libcall0d)(CMEM(cfa+CELL)))();
    *--sp=(TCELL)hlp;
    tos=hlp>>32;
  }
  NEXT;
xtcalli0:
  (*(libcalli0)(CMEM(cfa+CELL)))(tos);
  POP;
  NEXT;
xtcallii:
  tos=(*(libcallii)(CMEM(cfa+CELL)))(tos);
  NEXT;
xtcallip:
  tos=(*(libcallip)(CMEM(cfa+CELL)))(tos)-memory;
  NEXT;
xtcalli2: { long long hlp;
    hlp=(*(libcallid)(CMEM(cfa+CELL)))(tos);
    *--sp=(TCELL)hlp;
    tos=hlp>>32;
  }
  NEXT;
xtcallp0:
  (*(libcall00)(CMEM(cfa+CELL)))((unsigned char *)tos-memory);
  POP;
  NEXT;
xtcallpi:
  tos=(*(libcallii)(CMEM(cfa+CELL)))((unsigned char *)tos-memory);
  NEXT;
xtcallpp:
  tos=(*(libcallip)(CMEM(cfa+CELL)))((unsigned char *)tos-memory)-memory;
  NEXT;
xtcallp2: { long long hlp;
    hlp=(*(libcallid)(CMEM(cfa+CELL)))((unsigned char *)tos-memory);
    *--sp=(TCELL)hlp;
    tos=hlp>>32;
  }
  NEXT;
xtcall20:
  (*(libcalld0)(CMEM(cfa+CELL)))((((long long)tos)<<32)|*sp++);
  POP;
  NEXT;
xtcall2i:
  tos=(*(libcalldi)(CMEM(cfa+CELL)))((((long long)tos)<<32)|*sp++);
  NEXT;
xtcall2p:
  tos=(*(libcalldp)(CMEM(cfa+CELL)))((((long long)tos)<<32)|*sp++)-memory;
  NEXT;
xtcall22: { long long hlp;
    hlp=(*(libcalldd)(CMEM(cfa+CELL)))((((long long)tos)<<32)|*sp++);
    *--sp=(TCELL)hlp;
    tos=hlp>>32;
  }
  NEXT;
xtcall: { int mask,i,ap; static int args[10]; long long dhlp;
    static int m10[4]= { 0,10,10,20};
    *--sp=tos;
    mask=CMEM(cfa+2*CELL);
    i=m10[help=mask&3]; ap=0;
    while(mask>>=2) {
      switch(mask&3) {
        case 1: args[ap++]=*sp++;
          break;
        case 2: args[ap++]=(TCELL)memory+*sp++;
          break;
        case 3: (*((long long *)(args+ap)))=(((long long)(sp[0]))<<32)|sp[1];
          ap+=2; sp+=2;
          break;
        default: break;
      }
      i++;
    }
    goto *(mcalls[i]);
mc00: (*(libcall00)(CMEM(cfa+CELL)))(); goto mcend;
mc10: (*(libcall10)(CMEM(cfa+CELL)))(args[0]); goto mcend;
mc20: (*(libcall20)(CMEM(cfa+CELL)))(args[0],args[1]); goto mcend;
mc30: (*(libcall30)(CMEM(cfa+CELL)))(args[0],args[1],args[2]); goto mcend;
mc40: (*(libcall40)(CMEM(cfa+CELL)))(args[0],args[1],args[2],args[3]); goto mcend;
mc50: (*(libcall50)(CMEM(cfa+CELL)))(args[0],args[1],args[2],args[3],args[4]); goto mcend;
mc60: (*(libcall60)(CMEM(cfa+CELL)))(args[0],args[1],args[2],args[3],args[4],args[5]); goto mcend;
mc70: (*(libcall70)(CMEM(cfa+CELL)))(args[0],args[1],args[2],args[3],args[4],args[5],args[6]); goto mcend;
mc80: (*(libcall80)(CMEM(cfa+CELL)))(args[0],args[1],args[2],args[3],args[4],args[5],args[6],args[7]); goto mcend;
mc90: (*(libcall90)(CMEM(cfa+CELL)))(args[0],args[1],args[2],args[3],args[4],args[5],args[6],args[7],args[8]); goto mcend;
mc01: tos=(*(libcall01)(CMEM(cfa+CELL)))(); goto mcend;
mc11: tos=(*(libcall11)(CMEM(cfa+CELL)))(args[0]); goto mcend;
mc21: tos=(*(libcall21)(CMEM(cfa+CELL)))(args[0],args[1]); goto mcend;
mc31: tos=(*(libcall31)(CMEM(cfa+CELL)))(args[0],args[1],args[2]); goto mcend;
mc41: tos=(*(libcall41)(CMEM(cfa+CELL)))(args[0],args[1],args[2],args[3]); goto mcend;
mc51: tos=(*(libcall51)(CMEM(cfa+CELL)))(args[0],args[1],args[2],args[3],args[4]); goto mcend;
mc61: tos=(*(libcall61)(CMEM(cfa+CELL)))(args[0],args[1],args[2],args[3],args[4],args[5]); goto mcend;
mc71: tos=(*(libcall71)(CMEM(cfa+CELL)))(args[0],args[1],args[2],args[3],args[4],args[5],args[6]); goto mcend;
mc81: tos=(*(libcall81)(CMEM(cfa+CELL)))(args[0],args[1],args[2],args[3],args[4],args[5],args[6],args[7]); goto mcend;
mc91: tos=(*(libcall91)(CMEM(cfa+CELL)))(args[0],args[1],args[2],args[3],args[4],args[5],args[6],args[7],args[8]); goto mcend;
mc02: dhlp=(*(libcall02)(CMEM(cfa+CELL)))(); goto mcend;
mc12: dhlp=(*(libcall12)(CMEM(cfa+CELL)))(args[0]); goto mcend;
mc22: dhlp=(*(libcall22)(CMEM(cfa+CELL)))(args[0],args[1]); goto mcend;
mc32: dhlp=(*(libcall32)(CMEM(cfa+CELL)))(args[0],args[1],args[2]); goto mcend;
mc42: dhlp=(*(libcall42)(CMEM(cfa+CELL)))(args[0],args[1],args[2],args[3]); goto mcend;
mc52: dhlp=(*(libcall52)(CMEM(cfa+CELL)))(args[0],args[1],args[2],args[3],args[4]); goto mcend;
mc62: dhlp=(*(libcall62)(CMEM(cfa+CELL)))(args[0],args[1],args[2],args[3],args[4],args[5]); goto mcend;
mc72: dhlp=(*(libcall72)(CMEM(cfa+CELL)))(args[0],args[1],args[2],args[3],args[4],args[5],args[6]); goto mcend;
mc82: dhlp=(*(libcall82)(CMEM(cfa+CELL)))(args[0],args[1],args[2],args[3],args[4],args[5],args[6],args[7]); goto mcend;
mc92: dhlp=(*(libcall92)(CMEM(cfa+CELL)))(args[0],args[1],args[2],args[3],args[4],args[5],args[6],args[7],args[8]); goto mcend;
mcend:
    switch(help) {
      case 0: POP; break;
      case 1: break;
      case 2: tos-=(TCELL)memory; break;
      case 3: *--sp=(TCELL)dhlp; tos=dhlp>>32; break;
      default: break;
    }
  }
  NEXT;
xtlocinit:
  *--rp=(TCELL)lp;
  lp=(void *)rp;
  rp-=WMEM(ip);
  for(help=WMEM(ip+sizeof(short));help>0;help--) {
    *--rp=tos;POP;
  }
  ip+=CELL;
  NEXT;
xtunparm:
  rp=(TCELL *)lp;
  lp=(void *)*rp++;
  NEXT;
xtlocalloc:
  rp=(TCELL *)((((int)rp)-tos)&(-4));
  tos=(TCELL)rp-(TCELL)memory;
  NEXT;
xtlit:
  PUSH;
  tos=CMEM(ip); ip+=CELL;
  NEXT;
xtexec:
  cfa=tos; /* help=cfa-5; help-=memory[help]&0x3f;
  printf("%c%c\n",memory[help],memory[help+1]); */
  POP;
  goto ROUTINES(CMEM(cfa));
xtperf:
  cfa=CMEM(tos);
  POP;
  goto ROUTINES(CMEM(cfa));
xtnoop:
  NEXT;
xtbra:
  ip=CMEM(ip);
  NEXT;
xtqbra:
  if(tos) { ip+=CELL; } else { ip=CMEM(ip); }
  POP;
  NEXT;
xti:
  PUSH;
  tos=rp[0]+rp[1];
  NEXT;
xtj:
  PUSH;
  tos=rp[3]+rp[4];
  NEXT;
xtk:
  PUSH;
  tos=rp[6]+rp[7];
  NEXT;
xtdrop:
  POP;
  NEXT;
xtdup:
  PUSH;
  NEXT;
xtswap:
  help=*sp;
  *sp=tos;
  tos=help;
  NEXT;
xtover:
  PUSH;
  tos=sp[1];
  NEXT;
xtrot:
  help=*sp; *sp=tos;
  tos=sp[1]; sp[1]=help;
  NEXT;
xtmrot:
  help=sp[1]; sp[1]=tos;
  tos=*sp;
  *sp=help;
  NEXT;
xttuck:
  sp--;
  *sp=sp[1];
  sp[1]=tos;
  NEXT;
xtpick:
  tos=sp[tos];
  NEXT;
xtnip:
  sp++;
  NEXT;
xtqdup:
  if(tos) { PUSH; }
  NEXT;
xttdrop:
  sp+=2;
  tos=*(sp-1);
  NEXT;
xttdup:
  sp-=2;
  sp[1]=tos;
  *sp=sp[2];
  NEXT;
xttswap:
  help=sp[1]; sp[1]=tos; tos=help;
  help=*sp; *sp=sp[2]; sp[2]=help;
  NEXT;
xttover:
  sp-=2;
  sp[1]=tos;
  *sp=sp[4];
  tos=sp[3];
  NEXT;
xttrot:
  help=sp[1]; sp[1]=tos; tos=sp[3]; sp[3]=help;
  help=sp[2]; sp[2]=*sp; *sp=sp[4]; sp[4]=help;
  NEXT;
xtthdrop:
  sp+=2; POP;
  NEXT;
xtfdrop:
  sp+=3; POP;
  NEXT;
xtspfetch:
  PUSH;
  tos=(unsigned char *)sp-memory;
  NEXT;
xtspstore:
  sp=(TCELL *)(memory+tos);
  POP;
  NEXT;
xtplus:
  tos+=*sp++;
  NEXT;
xtminus:
  tos=(*sp++)-tos;
  NEXT;
xtunderp:
  sp[1]+=tos;
  POP;
  NEXT;
xtnegate:
  tos=-tos;
  NEXT;
xtabs:
  if(tos<0) tos=-tos;
  NEXT;
xttimes:
  tos*=*sp++;
  NEXT;
xtumt:
  { unsigned long long dhlp=(unsigned long long)tos*((unsigned long)*sp);
    *sp=(TCELL) dhlp;
    tos=(TCELL)(dhlp>>(8*CELL));
  }
  NEXT;
xtmt:
  { long long dhlp=(long long)tos*((long)*sp);
    *sp=(TCELL) dhlp;
    tos=(TCELL)(dhlp>>(8*CELL));
  }
  NEXT;
xtdiv:
  tos=((int)(*sp++))/((int)tos);
  NEXT;
xtumdmod:
  { unsigned long long dhlp=(((unsigned long long)(sp[0]))<<(8*CELL))+sp[1];
    sp[1]=dhlp%tos;
    sp++;
    tos=dhlp/tos;
  }
  NEXT;
xtsmdrem:
xtfmdmod:
xtdmod:
  help=((int)*sp)%((int)tos);
  tos=((int)*sp)/((int)tos);
  *sp=help;
  NEXT;
xttdmod:
xtdpl:
  { long long dhlp =(((long long)(sp[1]))<<(8*CELL))+sp[2];
    dhlp+=(((long long)(tos))<<(8*CELL))+*sp;
    sp+=2;
    *sp=(TCELL) dhlp;
    tos=(TCELL)(dhlp>>(8*CELL));
  }
  NEXT;
xtdmi:
  { long long dhlp =(((long long)(sp[1]))<<(8*CELL))+sp[2];
    dhlp-=(((long long)(tos))<<(8*CELL))+*sp;
    sp+=2;
    *sp=(TCELL) dhlp;
    tos=(TCELL)(dhlp>>(8*CELL));
  }
  NEXT;
xtdneg:
  { long long dhlp =(((long long)(tos))<<(8*CELL))+*sp;
    dhlp=-dhlp;
    *sp=(TCELL) dhlp;
    tos=(TCELL)(dhlp>>(8*CELL));
  }
  NEXT;
xtdabs:
  { long long dhlp =(((long long)(tos))<<(8*CELL))+*sp;
    if(dhlp<0) dhlp=-dhlp;
    *sp=(TCELL) dhlp;
    tos=(TCELL)(dhlp>>(8*CELL));
  }
  NEXT;
xtstd:
  PUSH;
  tos=-((int)tos<0);
  NEXT;
xttor:
  *--rp=tos;
  POP;
  NEXT;
xtfromr:
  PUSH;
  tos=*rp++;
  NEXT;
xtrfetch:
  PUSH;
  tos=*rp;
  NEXT;
xtttor:
  *--rp=*sp++;
  *--rp=tos;
  POP;
  NEXT;
xttfromr:
  PUSH;
  tos=*rp++;
  *--sp=*rp++;
  NEXT;
xttrfetch:
  PUSH;
  *--sp=rp[1];
  tos=*rp;
  NEXT;
xtrpfetch:
  PUSH;
  tos=(unsigned char *)rp-memory;
  NEXT;
xtrpstore:
  rp=(TCELL *)(memory+tos);
  POP;
  NEXT;
xtduptor:
  *(--rp)=tos;
  NEXT;
xtrdrop:
  rp++;
  NEXT;
xtstore:
  CMEM(tos)=*sp++;
  POP;
  NEXT;
xtfetch:
  tos=CMEM(tos);
  NEXT;
xtpstore:
  CMEM(tos)+=*sp++;
  POP;
  NEXT;
xtcstore:
  CHMEM(tos)=(char)(*sp++);
  POP;
  NEXT;
xtcfetch:
  tos=(TCELL)(CHMEM(tos));
  NEXT;
xtcpstore:
  CHMEM(tos)+=(char)(*sp++);
  POP;
  NEXT;
xtwstore:
  WMEM(tos)=(unsigned short)(*sp++);
  POP;
  NEXT;
xtwfetch:
  tos=(TCELL)(WMEM(tos));
  NEXT;
xtwpstore:
  WMEM(tos)+=(unsigned short)(*sp++);
  POP;
  NEXT;
xttstore:
  CMEM(tos)=*sp++;
  CMEM(tos+CELL)=*sp++;
  POP;
  NEXT;
xttfetch:
  *--sp=CMEM(tos+CELL);
  tos=CMEM(tos);
  NEXT;
xtpstoreplus:
  help=*sp++;
  CMEM(help)=tos;
  tos=help+CELL;
  NEXT;
xtwrdspl:
xtwrdjoin:
xtcnt:
  help=memory[tos++];
  PUSH;
  tos=help;
  NEXT;
xtwcnt:
  help=(TCELL)WMEM(tos);
  tos+=sizeof(short);
  PUSH;
  tos=help;
  NEXT;
xtlcnt:
  help=CMEM(tos);
  tos+=CELL;
  PUSH;
  tos=help;
  NEXT;
xtzcnt:
  PUSH;
  tos=strlen(memory+tos);
  NEXT;
xtopl:
  tos++;
  NEXT;
xtomi:
  tos--;
  NEXT;
xtcpl:
  tos+=CELL;
  NEXT;
xtcmi:
  tos-=CELL;
  NEXT;
xtcs:
  tos*=CELL;
  NEXT;
xtcspl:
  tos=(*sp++)+tos*CELL;
  NEXT;
xtcsmi:
  tos=(*sp++)-tos*CELL;
  NEXT;
xtplcs:
  tos+=(*sp++)*CELL;
  NEXT;
xtmics:
  tos-=(*sp++)*CELL;
  NEXT;
xtttimes:
  tos*=2;
  NEXT;
xttdiv:
  tos=((int)tos)/2;
  NEXT;
xtutdiv:
  tos=((unsigned int)tos)/2;
  NEXT;
xtchspl:
  tos+=*sp++;
  NEXT;
xtdttim:
xtdtdiv:
xtdo:
  *--rp=help=*(sp++)+0x80000000;
  *--rp=tos-help;
  POP;
  NEXT;
xtqdo:
  if(tos==*sp) { ip=CMEM(ip); }
  else {
    ip+=CELL;
    *--rp=help=*sp+0x80000000;
    *--rp=tos-help;
  }
  sp++; POP;
  NEXT;
xtloop:
  cfa=*rp;
  help=*rp=cfa+1;
  if(((signed long)(help^cfa))<0) {
    ip+=CELL;
    rp+=2;
  } else {
    ip=CMEM(ip);
  }
  NEXT;
xtploop:
  cfa=*rp;
  help=*rp=cfa+tos;
  POP;
  if(((signed long)(help^cfa))<0) {
    ip+=CELL;
    rp+=2;
  } else {
    ip=CMEM(ip);
  }
  NEXT;
xtof:
  if(*sp==tos) {
    sp++;
    ip+=CELL;
  } else {
    ip=CMEM(ip);
  }
  POP;
  NEXT;
xtunloop:
  rp+=2;
  NEXT;
xtmove:
  memmove(memory+*sp,memory+sp[1],tos);
  sp+=2;
  POP;
  NEXT;
xtcmove:
  for(tos+=cfa=sp[1],help=*sp;cfa<tos;cfa++,help++) {
    CHMEM(help)=CHMEM(cfa);
  }
  sp+=2; POP;
  NEXT;
xtcmove2:
  tos--;
  help=*sp+tos;
  for(tos+=cfa=sp[1];tos>=cfa;tos--,help--) {
    CHMEM(help)=CHMEM(tos);
  }
  sp+=2; POP;
  NEXT;
xtfill:
  memset(memory+sp[1],tos,*sp);
  sp+=2;
  POP;
  NEXT;
xtand:
  tos&=*sp++;
  NEXT;
xtor:
  tos|=*sp++;
  NEXT;
xtxor:
  tos^=*sp++;
  NEXT;
xtinv:
  tos=~tos;
  NEXT;
xtlsh:
  tos=(*(sp++))<<tos;
  NEXT;
xtrsh:
  tos=(*(sp++))>>tos;
  NEXT;
xtincr:
  (CMEM(tos))++;
  POP;
  NEXT;
xtdecr:
  (CMEM(tos))--;
  POP;
  NEXT;
xtcincr:
  CHMEM(tos)++;
  POP;
  NEXT;
xtcdecr:
  CHMEM(tos)--;
  POP;
  NEXT;
xttoggle:
xtzeq:
  tos=-(tos==0);
  NEXT;
xtzneq:
  tos=-(tos!=0);
xtzlss:
  tos=-((int)tos<0);
  NEXT;
xtzgrt:
  tos=-((int)tos>0);
  NEXT;
xteq:
  tos=-(*sp==tos);
  sp++;
  NEXT;
xtneq:
  tos=-(*sp!=tos);
  sp++;
  NEXT;
xtlss:
  tos=-((int)*sp<(int)tos);
  sp++;
  NEXT;
xtgrt:
  tos=-((int)*sp>(int)tos);
  sp++;
  NEXT;
xtulss:
  tos=-((unsigned int)*sp<(unsigned int)tos);
  sp++;
  NEXT;
xtugrt:
  tos=-((unsigned int)*sp>(unsigned int)tos);
  sp++;
  NEXT;
xtdulss:
xtbetween:
  sp+=2;
  tos=(sp[-1]<=tos && sp[-1]>=sp[-2])?-1:0;
  NEXT;
xtwithin:
  sp+=2;
  tos=(sp[-1]<tos && sp[-1]>=sp[-2])?-1:0;
  NEXT;
xttrailch:
xtcutstr:
  tos=(*sp>tos)?tos:*sp;
  sp[1]+=tos;
  tos=(*sp++)-tos;
  NEXT;
xtuemit:
  putchar(tos);
  POP;
  NEXT;
xtutype:
  fwrite(memory+*sp++,1,tos,stdout);
  POP;
  NEXT;
xtubye:
  return(0);
xtucr:
  putchar('\n');
  NEXT;
xtugotoxy:
xtukey:
  PUSH
  tos=getchar();
  NEXT;
xtukeyq:
  PUSH;
  tos=-1;
  NEXT;
xtopfi:
  memcpy(tbuf,memory+sp[1],*sp);
  tbuf[*sp++]=0;
  tos=-((*sp=(TCELL)fopen(tbuf,memory+tos))==(int)NULL);
  NEXT;
xtcrfi:
  memcpy(tbuf,memory+sp[1],*sp);
  tbuf[*sp++]=0;
  tos=-((*sp=(TCELL)fopen(tbuf,memory+tos))==(int)NULL);
  if(tos!=0) tos=fclose((FILE *)*sp);
  NEXT;
xtclfi:
  tos=fclose((FILE *)tos);
  NEXT;
xtrdfi:
  *++sp=fread(memory+sp[1],1,*sp,(FILE *)tos);
  tos=0;
  NEXT;
xtwrfi:
  tos=-(fwrite(memory+sp[1],1,*sp,(FILE *)tos)!=*sp);
  sp+=2;
  NEXT;
xtposfi:
xtrpfi: { fpos_t pos;
  pos=((fpos_t)sp[1])+(((fpos_t)sp[0])<<32);
  tos=fsetpos((FILE *)tos,&pos); /* double cell order ? */
  sp+=2;
  }
  NEXT;
xtfipos: { fpos_t pos;
  sp-=2;
  tos=fgetpos((FILE *)tos,&pos);
  sp[1]=pos; sp[0]=pos>>32;
  }
  NEXT;
xtprswht:
  help=*sp;
  while(tos) {
    if(((unsigned char)(memory[help]))>' ') break;
    help++; tos--;
  }
  cfa=help;
  while(tos) {
    if(((unsigned char)(memory[help]))<=' ') break;
    help++; tos--;
  }
  *sp=cfa;
  tos=help-cfa;
  NEXT;
xtcomp:
  if((help=sp[1])>tos) help=tos;
  if((cfa=memcmp(memory+sp[2],memory+*sp,help))!=0) {
    tos=(cfa>0)*2-1;
  } else {
    if(sp[1]>tos) { tos=1; }
    else {
      if(sp[1]<tos) { tos=-1; }
      else tos=0;
    }
  }
  sp+=3;
  NEXT;
xtccomp:
  { unsigned char *str1,*str2;
    if((help=sp[1])<tos) { cfa=-1; } else { cfa=(help>tos); help=tos; }
    str1=memory+sp[2]; str2=memory+sp[0];
    for(;help>0;help--,str1++,str2++) {
      if(MYUP(*str1)!=*str2) {
        cfa=2*(MYUP(*str1)>*str2)-1;
        break;
      }
    }
  }
  tos=cfa;
  sp+=3;
  NEXT;
xthash:
  cfa=(int)(memory+sp[1]); sp[1]=tos; tos=help=*sp; sp+=2;
  while(help!=0) {
    tos=(tos<<7)|(tos>>25);
    tos+=MYUP(*((unsigned char *)cfa));
    cfa++;help--;
  }
  tos=((unsigned int)tos%*((unsigned int *)(sp-1)))*CELL;
  NEXT;
xtallocate:
  if((tos=(TCELL)malloc(tos))==(TCELL)NULL)
    { tos=0; } else { tos-=(TCELL)memory; }
  NEXT;
xtresize:
  if((tos=(TCELL)realloc(memory+*sp,tos))==(TCELL)NULL) {
    tos=0;
  } else {
    tos-=(TCELL)memory;
  }
  sp++;
  NEXT;
xtfree:
  free(memory+tos);
  POP;
  NEXT;
xtnextimage:
  tos=get_next_image_info(tos,(int32 *)(memory+*sp),(image_info *)(memory+sp[1]));
  sp+=2;
  NEXT;
xtimgsymb:
  tos=get_image_symbol(tos,(char *)(memory+sp[0]),(int32)(sp[1]),(void **)(memory+sp[2]));
  sp+=3;
  NEXT;
xtloadaddon:
  tos=(TCELL)load_add_on((char *)(memory+tos));
  NEXT;
xtabs2rel:
  tos=(unsigned char *)tos-memory;
  NEXT;
xtrel2abs:
  tos=(TCELL)(memory+tos);
  NEXT;
}

int main(int argc,char **argv)
{ long info[7];
  char *area;
  FILE *fin;
  int i;
  struct termios term;
  char name[256];

  area=argv[0]+strlen(argv[0]);
  while(--area>=argv[0]&&*area!='/');
  i=area[1];  area[1]='\0';
  strcpy(name,argv[0]);
  strcat(name,area+2);
  area[1]=i; 
  if((fin=fopen(name,"rb"))==NULL) {
    fprintf(stderr,"Image file \"%s\" not found\n",name);
    exit(1);
  }        
  fread(info,sizeof(long),7,fin);
  area=(unsigned char *)malloc(info[2]);
  memset(area,0,info[2]);
  fread(area+7*sizeof(long),1,info[2]-7*sizeof(long),fin);
  fclose(fin);
  memcpy(area,info,5*sizeof(long));
  
  ((TCELL *)area)[5]=argc;
  ((TCELL *)area)[6]=(TCELL)((char *)(argv)-area);
  for(i=0;i<argc;i++) argv[i]-=(int)area;

  fflush(stdin);
  setvbuf(stdin,NULL,_IONBF,0);
  fflush(stdout);
  setvbuf(stdout,NULL,_IONBF,0);
  tcgetattr(STDIN_FILENO,&term);
  i=term.c_lflag;
  term.c_lflag&=~(ECHO|ECHONL|ICANON|ISIG|IEXTEN);
  tcsetattr(STDIN_FILENO,TCSANOW,&term);
  
  machine(area);

  tcgetattr(STDIN_FILENO,&term);
  term.c_lflag=i;
  tcsetattr(STDIN_FILENO,TCSANOW,&term);
  fflush(stdout);
  setvbuf(stdout,NULL,_IOLBF,0);
  fflush(stdin);
  setvbuf(stdin,NULL,_IOLBF,0);

  for(i=0;i<argc;i++) argv[i]+=(int)area;
  free(area);
  return(0);
}